////////////////////////////////////////////////////////////////////////////////////////////////
//   dictionaries.js
//
//   Author: Erwan Loisant
//   Date: September 2003
//   Licence: General Public Licence
//   Description: The dictionary class manages load and
//                access to dictionary files from disk
//
//   Contributors: Todd Rudick, Mathieu Mangeot
////////////////////////////////////////////////////////////////////////////////////////////////

//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

const NS_SEEK_SET = Components.interfaces.nsISeekableStream.NS_SEEK_SET;
const NS_SEEK_CUR = Components.interfaces.nsISeekableStream.NS_SEEK_CUR;
const NS_SEEK_END = Components.interfaces.nsISeekableStream.NS_SEEK_END;

var strbundle = window.document.getElementById("strings");

function File(aPath)
{
  this.path = aPath;

  this.uri = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIURI);
  this.uri.spec = aPath;

  this.fios = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
}

File.prototype = {
  path : null,
  uri : null,
  fios : null,
  istream : null,
  scstream : null,
  chan : null,
  iseof : false,
  tellval : 0,
  usetellval : false,
  isreadline : null,
    
  open : function ()
  {
    this.chan = this.fios.newChannelFromURI(this.uri);
    this.istream = this.chan.open();
    this.scstream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
    this.scstream.init(this.istream);
    
    this.iseof = (this.istream.available() == 0);
    this.tellval = 0;
    this.usetellval = false;
    this.isreadline = null;
  },
  
  canReadLine : function()
  {
    if ( this.isreadline == null ) {
      this.isreadline = this.istream instanceof Components.interfaces.nsILineInputStream;
    }
    
    return this.isreadline;
  },
  
  readLine : function ()
  {
    if ( this.canReadLine() ) {
      var buf = {};
      
      this.iseof = ! this.istream.readLine(buf);
      
      tellval = tellval + buf.value.length;
      
      // add a byte for the newline at the end which readLine does not return.
      if ( ! this.iseof )
        tellval += 1;
        
      usetellval = true;
      
      return buf.value;
    } else
      throw new Exception("readLine() not available on stream to " + this.path);
  },
  
  read : function ( count ) 
  {
    if ( usetellval ) {
      // we have to adjust the current seek position on the stream as readLine did a buffered
      // read before and the current stream position is not the next byte to be read...
      this.seek(NS_SEEK_SET, tellval);
    }
    
    var s = this.scstream.read(count);
    
    this.iseof = ! ( this.istream.available() > 0 );
    
    return s;
  },

  canSeek : function()
  {
    return (this.istream instanceof Components.interfaces.nsISeekableStream);
  },
  
  seek : function ( whence, offset ) 
  {
    if ( this.canSeek() ) {
      if ( whence == NS_SEEK_CUR ) {
        if ( usetellval ) {
          offset = tellval + offset;
          this.istream.seek(NS_SEEK_SET, offset);
        } else
          this.istream.seek(whence, offset);
      } else 
        this.istream.seek(whence, offset);
      
      tellval = this.istream.tell();
      usetellval = false;
    } else
      throw new Exception("seek() not available on stream to " + this.path);
  },
  
  tell : function()
  {
    if ( this.canSeek() ) {
      if ( usetellval )
        return tellval;
      else {
        tellval = this.istream.tell();
        return tellval;
      }
    } else 
      throw new Exception("tell() not available on stream to " + this.path);
  },
  
  getSize : function ()
  {
    if ( this.canSeek() ) {
      var curpos = this.istream.tell();
      this.seek(NS_SEEK_END, 0);
      var size = this.istream.tell();
      this.seek(NS_SEEK_SET, curpos);
      return size;
    } else
      throw new Exception("tell() for getSize() not available on stream to " + this.path);
  },
  
  eof : function ()
  {
    return this.iseof;
  },
  
  close : function ()
  {
    this.istream.close();
  }
}

function IdxEntry(word, pos, rpos)
{
  this.word = word;
  this.pos = pos;
  this.rpos = rpos;
  this.left = null;
  this.right = null;
  this.searched = false;
}

IdxEntry.prototype = {
  word : null,
  pos : null,
  rpos : null,
  left : null,
  right : null,
  searched : null
}

function DictionaryFile(path)
{
  this.path = path;
  
  this.dictfile = new File(this.path);
  this.dictfile.open();
  
  this.dictsize = this.dictfile.getSize();
  
  this.uniconv = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
  this.uniconv.charset = "UTF-8";
  this.searchindex = new Array();
  
  // alert("checking " + path);
  // debugger; 
  // 
  // var lw = "";
  // var lc = 0;
  // var np = 10;
  // 
  // while ( ! this.dictfile.eof() ) {
  //  var l = this.dictfile.readLine();
  //  l = this.uniconv.ConvertToUnicode(l);
  //  lc = lc + 1;
  //  var w = l.split(" ")[0];
  //  
  //  if ( w < lw ) {
  //    debugger;
  //  }
  //   
  //  lw = w;
  //  
  //  if ( this.dictfile.tell() * 100 / this.dictsize > np ) {
  //    alert("checked " + np + "%");
  //    np = np + 10;
  //  }
  // }
  
  // initialize the anker for the searchindex. It is the entry that starts right after the middle of the file
  this.dictfile.seek(NS_SEEK_SET, this.dictsize / 2);
  this.dictfile.readLine();

  var pos = this.dictfile.tell();
  var entry = this.dictfile.readLine();
  var rpos = this.dictfile.tell();
  entry = this.uniconv.ConvertToUnicode(entry);
  var sp = entry.split(this.fieldsep);
  var idxentry = new IdxEntry(sp[0], pos, rpos);
  this.searchindex.push(idxentry);
}

DictionaryFile.prototype = {
  path : null,
  dictfile : null,
  dictsize : null,
  searchindex : null,
  uniconv : null,
  fieldsep : ' ',
  
  close : function()
  {
    this.dictfile.close();
  },
  
  search : function(word)
  {
    var si = this._search(word, 0, null, null);
    
    if ( si >= 0 )
    {
      this.dictfile.seek(NS_SEEK_SET, this.searchindex[si].pos);
      var entries = new Array();
      
      while ( true ) {
        var entry = this.dictfile.readLine();
        entry = this.uniconv.ConvertToUnicode(entry);
        var sp = entry.split(this.fieldsep);

        if ( sp[0] == word )
          entries.push(entry);
        else
          break;
      };
      
      return entries;
    } else
      return null;
  },

  // Search for word at search index entry six within boundaries if entries lix and rix.
  _search : function(word, six, lix, rix)
  {
    if ( word <= this.searchindex[six].word ) {
      // search in the left half
      
      var nix = null;
      
      if ( this.searchindex[six].left == null ) {
        // the left branch has not been searched, yet...
        
        var lixrpos;
        
        if ( lix == null )
          lixrpos = 0;
        else
          lixrpos = this.searchindex[lix].rpos;
          
        // if lix ends where six begins there is nothing left between those entries...
        if ( lixrpos == this.searchindex[six].pos ) {
          this.searchindex[six].left = -1;
          if ( word == this.searchindex[six].word ) {
            this.searchindex[six].searched = true;
            return six;
          } else {
            return -1;
          }
        }

        var spos;
        spos = ( lixrpos + this.searchindex[six].pos ) / 2;
          
        this.dictfile.seek(NS_SEEK_SET, spos);
        this.dictfile.readLine();

        var pos = this.dictfile.tell();
        var entry = null;
        
        if ( pos == this.searchindex[six].pos ) {
          this.dictfile.seek(NS_SEEK_SET, lixrpos);
          var tpos = lixrpos;
          
          while ( tpos < this.searchindex[six].pos ) {
            pos = tpos;
            entry = this.dictfile.readLine();
            tpos = this.dictfile.tell();
          }
        }

        if ( entry == null )        
          entry = this.dictfile.readLine();
          
        var rpos = this.dictfile.tell();
        entry = this.uniconv.ConvertToUnicode(entry);
        var sp = entry.split(this.fieldsep);
        
        var idxentry = new IdxEntry(sp[0], pos, rpos);
        nix = this.searchindex.length;
        this.searchindex.push(idxentry);
        
        this.searchindex[six].left = nix;
      } else if ( this.searchindex[six].left == -1 ) {
        if ( word == this.searchindex[six].word )
          return six;
        else 
          return -1;
      } else {
        nix = this.searchindex[six].left;
      }
      
      if ( ( word == this.searchindex[six].word ) && this.searchindex[six].searched ) 
        return six;
      
      var retval = this._search(word, nix, lix, six);
      
      if ( ( retval == -1 ) && ( word == this.searchindex[six].word ) ) {
        this.searchindex[six].searched = true;
        return six;
      } else {
        return retval;
      }
    } else if ( word > this.searchindex[six].word ) {
      // search in the right half
      
      var nix = null;
        
      if ( this.searchindex[six].right == null ) {
        // the right branch has not been searched, yet...
        
        var spos;
        if ( rix == null )
          spos = this.dictsize;
        else
          spos = this.searchindex[rix].pos;
                  
        // if six ends where rix begins, there is nothing left inbetween...
        if ( this.searchindex[six].rpos == spos ) {
          this.searchindex[six].right = -1;
          return -1;
        }
        
        spos = ( this.searchindex[six].rpos + spos ) / 2;
          
        this.dictfile.seek(NS_SEEK_SET, spos);
        this.dictfile.readLine();

        var pos = this.dictfile.tell();
        var entry = null;
        var reset = false;
        var rpos;
        
        if ( rix == null )
          rpos = this.dictsize;
        else
          rpos = this.searchindex[rix].pos;
          
        // O.K. Here we have to check whether or not we have actually reached the rix border.
        // This is a little bit trick here if rix is null and the definite border is defined
        // by EOF. However, there may be a NL before the last entry which basically means that 
        // there is an (empty) line right before the EOF.
        if ( ! this.dictfile.iseof && ( pos < rpos  ) )
        {
          entry = this.dictfile.readLine();
          
          if ( this.dictfile.iseof && ( entry.length == 0 ) )
            reset = true;
        } else
          reset = true;
        
        if ( reset ) {
          // there is no entry starting after spos. Thus we start looking from six...
          this.dictfile.seek(NS_SEEK_SET, this.searchindex[six].rpos);
          var tpos = this.searchindex[six].rpos;
          entry = null;
          
          while ( ! this.dictfile.iseof && ( tpos < rpos ) ) {
            var t = this.dictfile.readLine();
            
            // TODO: think about empty lines in the dictionary...
            if ( ( t.length > 0 ) || ! this.dictfile.iseof ) {
              pos = tpos;
              entry = t;
              tpos = this.dictfile.tell();
            }
          }
          
          // nothing found.
          if ( entry == null ) {
            this.searchindex[six].right = -1;
            return -1;
          }
        } 
        
        rpos = this.dictfile.tell();
        entry = this.uniconv.ConvertToUnicode(entry);
        var sp = entry.split(this.fieldsep);
        var idxentry = new IdxEntry(sp[0], pos, rpos);
        nix = this.searchindex.length;
        this.searchindex.push(idxentry);
        
        this.searchindex[six].right = nix;
      } else if ( this.searchindex[six].right == -1 ) {
        return -1;
      } else {
        nix = this.searchindex[six].right;
      }
      
      return this._search(word, nix, six, rix);
    } 
  }
}

function EDict(name)
{
  this.name = name;
  this.chromepath = "chrome://" + name + "/content/";
  this.dictfile = new DictionaryFile(this.chromepath + "edict");
  
  try {
    this.vconj = new VConj(this.chromepath + "vconj");
  } catch (e) {
    this.vconj = null; 
  }
}

EDict.prototype = {
  chromepath : null,
  name : null,
  dictfile : null,
  vconj : null,
  
  // direct search into the dictionary file.
  _search : function(word) {
    return this.dictfile.search(word);
  },
  
  // search for word and conjugations.
  // returns an error of matching dictionary entries including conjugation reason if applicable.
  search : function(word) {
    var tries = Array();
    var gotcha = Array();
    
    if ( this.vconj == null )
    {
      var o = new Object();
      o.txt = word;
      
      tries.push(o);
    } else {
      tries = this.vconj.kanjiConjugate(word);
    }
    
    var entries = Array();
    
    for ( var i = 0; i < tries.length; i++ ) {
      if ( gotcha[tries[i].txt] ) 
        continue;
        
      gotcha[tries[i].txt] = true;
      
      var found = this.dictfile.search( tries[i].txt );
      
      if ( found ) {
        for ( var j = 0; j < found.length; j++ ) {
          var o = new Object();
          
          o.txt = found[j];
          
          if ( tries[i].reason ) 
            o.reason = tries[i].reason;
            
          entries.push(o);
        }
      }
    }
    
    if ( entries.length == 0 )
      return null;
    else
      return entries;
  },
  
  close : function() {
    this.dictfile.close();
  }
}

function KanjiDict(name)
{
  this.name = name;
  this.chromepath = "chrome://" + name + "/content/";
  this.dictfile = new DictionaryFile(this.chromepath + "kanjidic");
}

KanjiDict.prototype = {
  chromepath : null,
  name : null,
  dictfile : null,
  radicals : null,
  
  search : function(word) {
    // deferred initialization of radicals file because it takes some time to load...
    if ( this.radicals == null ) 
      this.radicals = new Radicals(this.chromepath);

    var found = this.dictfile.search(word);
    var entries = Array();
    
    if ( found ) {
      for ( var i = 0; i < found.length; i++ ) {
        var o = new Object();
        o.txt = found[i];
        this._analyze(o);
        entries.push(o);
      } 
      
      return entries;
    } else 
      return null;
  },
  
  _analyze : function(entry) {
    var txt = entry.txt;
    
    // split off the meanings first. meanings are text in {} but may contain spaces which are field
    // separators anywhere else.
    entry.meanings = txt.match(/{[^}]*}/g);
    
    // remove the braces.
    for ( var i = 0; i < entry.meanings.length; i++ ) {
      entry.meanings[i] = entry.meanings[i].replace(/{(.*)}/, "$1");
    }
    
    // remove them from the dict entry
    txt = txt.replace(/{[^}]*}/g, "");
    // this leaves multiple spaces in the entry which we condense to a single space here.
    txt = txt.replace(/  +/g, " ");
    
    var fields = txt.split(" ");
    
    entry.key = fields[0];
    entry.jiscode = parseInt(fields[1], 16);
    entry.reading = new Array();
    entry.nameReading = new Array();
    entry.radicalReading = new Array();
    entry.info = new Object();
    
    var readingtype = 0;
    
    for ( var i = 2; i < fields.length; i++ ) {
      if ( fields[i].length == 0 )
        continue;
        
      var k = null;
      var v = null;
      
      if ( ( fields[i].charAt(0) == 'M' ) 
          || ( fields[i].charAt(0) == 'D' ) 
          || ( fields[i].substring(0, 2) == 'IN' ) ) {
        k = fields[i].substring(0, 2);
        v = fields[i].substring(2);
      } else if ( fields[i].match(/^[A-Z]/) ) {
        k = fields[i].substring(0, 1);
        v = fields[i].substring(1);
      } 
      
      if ( k == null ) {
        if ( readingtype == 0 ) {
          entry.reading.push(fields[i]);
        } else if ( readingtype == 1 ) {
          entry.nameReading.push(fields[i]);
        } else if ( readingtype == 2 ) {
          entry.radicalReading.push(fields[i]);
        }
      } else {
        // name readings and radical readings are prefixed with field T1 or T2. There seem to be
        // several readings allowed after each of these prefixes. I don't know whether or not 
        // there may be some normal readings coming sometime later and I don't think so but
        // as Breen wrote in the docs: "which may be in any order and interspersed" so I 
        // assume we have to reset the readingtype once we have found something else which is
        // however not fully correct as we have already removed the meanings... 
        readingtype = 0;
        
        if ( k == "T" ) {
          if ( v == "1" )
            readingtype = 1;
          else if ( v == "2" ) 
            readingtype = 2;
          else
            throw "Unknown reading type " + fields[i];
        } else {
          // anything else is additional info that is store with key and value.
          // we don't need any more interpretation of the detailed contents as
          // we basically just copy the information into the output...
          
          if ( entry.info[k] )
            entry.info[k] += " " + v;
          else
            entry.info[k] = v;
        }
      }
    }
  },
  
  close : function() {
    this.dictfile.close();
  }
}


function VConj(path)
{
  this.path = path;
  this.vconjDescription = new Object();
  this.uniConv = new ucConverter();
  
  this.load(path);
}

VConj.prototype = {
  // Instance variables
  path : null,
  vconjData : null, 
  vconjDescription : null,
  uniConv : null,
  encoding : "EUC-JP",

  //////////////////////////////////////////////////////////////
  // loadVConj
  //
  // PARAMETERS:
  //   path: string
  //     absolute file path of radical file
  //   encoding: string
  //     encoding of radical file
  // RETURNS:
  //   nothing
  //
  // Load conjugasons file giving plain form from a given verbal form.
  // prepare global variables vconjData and vconjDescription.
  // Initial Developer: Todd Rudick (http://www.rikai.com)
  //////////////////////////////////////////////////////////////
  load : function(path) {
    this.vconjData = new Array();
    
    var f = new File(path);
    
    try {
      f.open();
    } catch ( e ) {
      alert(strbundle.getString("vconjopenerror") + " " + path);
      throw e;
    }
    
    while ( ! f.eof() ) {
      var line = this.uniConv.toUnicode( f.readLine(), this.encoding );
      
      if ( line.match(/^#/) ) {
        continue;
      } else if ( line.charAt(0)=='$' ) {
        break;
      } else {
        var fields = line.split("\t");
        this.vconjDescription[fields[0]] = fields[1];
      }
    }
    
    // conjugations from here
    while ( ! f.eof() ) {
      var line = this.uniConv.toUnicode( f.readLine(), this.encoding );
      
      if (line.match(/^#/)) {
        continue;
      } else {
        var fields = line.split("\t");
        var def = new Object();
        def.before = fields[0];
        def.after = fields[1];
        def.reason = fields[2];
        this.vconjData.push(def);
      } // if ... else
    }
    
    f.close();
  }, // laodVConj()

  //////////////////////////////////////////////////////////////
  // kanjiConjugate
  //
  // PARAMETERS:
  //   txt: string
  // RETURNS: array(txt, reason)
  //
  // try to conjugate txt into other possibilities, return array 
  // of object(txt,reason). no reason for the original txt, which is returned.
  // the reason is a numeric code meaning eg "past" or "te form".
  // See vconj file for more details.
  //
  // Initial Developer: Todd Rudick (http://www.rikai.com)
  //////////////////////////////////////////////////////////////
  kanjiConjugate : function(txt) {
    var ret = new Array();
    ret.push( new Object() );
    ret[0].txt = txt;
    
    var wordHash = new Object();
    wordHash[txt] = true;
    
    for ( var i=0; i < this.vconjData.length;i++) {    
      var vconjDataElement = this.vconjData[i];

      if ( ! vconjDataElement.before )
        continue;
        
      var olen = ret.length;
      
      for (var w = 0; w < olen; w++) {
        var ind = ret[w].txt.lastIndexOf(vconjDataElement.before);
        
        if ( ind != -1 ) {
          if ( ind == ret[w].txt.length - vconjDataElement.before.length ) {
	        // Create a new (txt, reason) to be added to the result
            var o = new Object();
            
            o.txt = ret[w].txt.substr(0,ind) + vconjDataElement.after;
            
            if ( !wordHash[o.txt] ) {
              if ( ret[w].reason ) {
                o.reason = ret[w].reason + ", " + this.vconjDescription[vconjDataElement.reason];
              } else {
                o.reason = this.vconjDescription[vconjDataElement.reason];
              }
	          //alert(o.txt + " - " + o.reason);
              ret.push(o);
            }
          }
        }
      }
    }
    return ret;
  }
}

function Radicals(path)
{
  this.path = path;
  this.uniConv = new ucConverter();
  
  this.loadRadicals(path + "radicals.tm", "EUC-JP");
  this.loadRadicalMeaning(path + "radical_meanings", "UTF-8");
  this.loadRadkfile(path + "radkfile", "EUC-JP");
}

Radicals.prototype = {
  // Instance variables
  path : null,
  radicalData : null, 
  radicalMeaning : null,
  radicalReading : null,
  radkfileData : null,
  uniConv : null,

  loadRadicals : function(path, encoding) {
    var f = new File(path);

    try {
      f.open();
    } catch ( e ) {
      alert(strbundle.getString("radfileopenerror") + " " + path);
      throw e;
    }

    this.radicalData = new Object();
    
    while (!f.iseof)
    {
      var line = this.uniConv.toUnicode(f.readLine(), encoding);
      
      var rad = line.match(/^([^ ]*) ([^ ]*).*/);
      
      if ( rad[1].charCodeAt(0) > 255 ) {
        if ( ( rad[2].charAt(0)=='[' )
          || ( rad[2].charAt(0)=='(' ) ) {
          var val = rad[2];
          val = val.match(/^.(.*).$/)[1];
          
          this.radicalData[rad[1]] = val; // use rad->num
        } else {
          this.radicalData[rad[1]] = rad[2];
          this.radicalData[rad[2]] = rad[1]; // do num->rad at the same time
        }
      }
    }
    
    f.close();
  },

  loadRadicalMeaning : function(path, encoding)
  {  
    var f2 = new File(path);
  
    try {
      f2.open();
    } catch ( e ) {
      alert(strbundle.getString("radmeanopenerror") + " " + path);
      throw e;
    }
  
    i=0;
    this.radicalMeaning = new Object();
    this.radicalReading = new Object();
  
    while (!f2.iseof)
    {
      var line = this.uniConv.toUnicode(f2.readLine(), encoding).split(" ");
      
      // TODO: figure out when this actually changes this.radicalData
      
      this.radicalData[line[0]] = String(i+1);
      this.radicalData[i+1] = line[0];
      this.radicalReading[i+1] = line[1];
      this.radicalMeaning[i+1] = line[2];
      this.radicalReading[line[0]] = line[1];
      this.radicalMeaning[line[0]] = line[2];
      
      
      i++;
    }
    
    f2.close();
    
    // TODO obsolete this code.
    // fix up radicals that aren't represented in proper form
    // for (var rad in this.radicalData) {
    //   if ( ! rad.match(/^[0-9]*$/) ) {
    //     if ( this.radicalData[rad].match(/^[0-9]*$/) ) {
    //       if ( !this.radicalData[this.radicalData[rad]] ) {
    //         debugger;
    //         this.radicalData[this.radicalData[rad]] = rad;
    //       }
    //     }
    //   }
    // }
  },
  
//////////////////////////////////////////////////////////////
// loadRadkfile
//
// PARAMETERS:
//   path: string
//     absolute file path of radical file
//   encoding: string
//     encoding of radical file
// RETURNS:
//   array (radkfileData)
//
// Initial Developer: Todd Rudick (http://www.rikai.com)
//////////////////////////////////////////////////////////////
  loadRadkfile : function(path, encoding)
  {
    var f = new File(path);
    
    try {
      f.open();
    } catch ( e ) {
      alert(strbundle.getString("radkfileopenerror") + " " + path);
      throw e;
    }
  
    this.radkfileData = new Object();
    var radical = "";
  
    while ( ! f.iseof )
    {
      var line = this.uniConv.toUnicode(f.readLine(), encoding);
    
      if (line.match(/^#/)) {
        continue;
      } else if ( line.match(/^\$/) ) {
        lineA = line.split(/ /);
        
        radical = lineA[1];
        
        if ( this.radicalData[radical] ) {
          if ( this.radicalData[this.radicalData[radical]] )
            radical = this.radicalData[this.radicalData[radical]];
        }
      } else {
        for (var j = 0; j < line.length; j++) {
          if ( line.charCodeAt(j) > 255 ) {
            var c = line.charAt(j);
            if ( ! this.radkfileData[c] ) {
              this.radkfileData[c] = new Object();
            }
            this.radkfileData[c][radical] = true;
          }
        }
      }
    }
    
    f.close();
  }
}

////////////////////////////////////////////////////////////////////////////////
// ucConverter -- a class to convert to/from unicode (hopefully)
//
// Based on MessageManager() from chatzilla, by Rob Ginda
// Used here under the terms of the GPL
////////////////////////////////////////////////////////////////////////////////
function ucConverter()
{
  const UC_CTRID = "@mozilla.org/intl/scriptableunicodeconverter";
  const nsIUnicodeConverter =
    Components.interfaces.nsIScriptableUnicodeConverter;

  this.ucConverter =
      Components.classes[UC_CTRID].getService(nsIUnicodeConverter);
  this.defaultBundle = null;
  this.bundleList = new Array();

  // Convert a string read from disk to unicode ; if no charset, unicode is used
  this.toUnicode = function (msg, charset) {
    if (!charset) charset = "UTF-8";
    try
    {
      this.ucConverter.charset = charset;
      msg = this.ucConverter.ConvertToUnicode(msg);
    }
    catch (ex)
    { alert ("caught exception " + ex + " converting '" + msg + "' to charset " + charset);}
    return msg;
  };

  // Convert a string from unicode to be written to disk ; if no charset, unicode is used
  this.fromUnicode = function (msg, charset) {
    if (!charset)
      charset = "UTF-8";
      //return msg;
    if (charset != this.ucConverter.charset)
      this.ucConverter.charset = charset;
    try
    {
      if ("Finish" in this.ucConverter)
      {
        msg = this.ucConverter.ConvertFromUnicode(msg);
        this.ucConverter.Finish();
      }
      else
      {
        msg = this.ucConverter.ConvertFromUnicode(msg + " ");
        msg = msg.substr(0, msg.length - 1);
      }
    }
    catch (ex)
      {alert ("caught exception " + ex + " converting '" + msg + "' from charset " + charset);}
    return msg;
  };
}